==> pam_bsdbioapi-1.5.tar.gz: pam_bsdbioapi-1.5/src/pam_bsdbioapi/pam_bsdbioapi.c <==
/*-
 * Copyright (c) 2006 Fredrik Lindberg. <fli at shapeshifter dot se>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: pam_bsdbioapi.c 57 2006-02-23 14:15:13Z fli $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/queue.h>
#include <login_cap.h>
#include <pwd.h>

#include <bioapi.h>
#include <birdb.h> 
#include <int_bioapi.h>

#define PAM_SM_AUTH
#define PAM_SM_PASSWORD

#include <security/pam_appl.h>
#include <security/pam_modules.h>
#include <security/pam_mod_misc.h>

#define DEFCONFPATH "/usr/local/etc/birdb.conf"

/*
 * Callback state, passed to bioapi_guicallback
 */
struct callbackh {
	pam_handle_t *pamh;
	char **msgs;
	int maxmsg;
	int prevmsg;
};

/*
 * Return position and length of next logical line (separated by \n)
 */
static int
nextline(FILE *fp, size_t *len)
{
	int ch;
	long tmp, newlen = 0;

	/* Save our file position */
	tmp = ftell(fp);

	/* Seek until we hit \n (end of line) or EOF */
	do {
#if __POSIX_VISIBLE >= 199506
		ch = getc_unlocked(fp);
#else
		ch = getc(fp);
#endif
		newlen++;
	} while ((ch != EOF) && ((char)ch != '\n'));

	/* Restore position to previous value */
	fseek(fp, tmp, SEEK_SET);

	*len = newlen;
	return (ch);
}

/*
 * Truncate string at the occurance of a '#'-character
 */
static void
commentclr(char *str)
{
	char *p = str;

	do {
		if (*p == '#') { *p = 0; }
	} while (*p++ != 0);
}

/*
 * Parse callback message files
 * Callers must cope with that the returned array may have holes in
 * it, the holes does always contain a NULL pointer.
 */
static int
parse_msgfile(const char *path, char **msg[])
{
	size_t len, linesz = 80, rlen, res;
	char *buf, *tmp = NULL, **ma = NULL;
	int msgs = 0, index, prev_index, high_index, i;
	FILE *fp;
    
	fp = fopen(path, "r");
	if (fp == NULL)
		return (-1);

	buf = malloc(linesz);
	if (buf == NULL)
		return (-1);

	prev_index = high_index = 0;
	ma = malloc(sizeof(char *));
	while (nextline(fp, &len) != EOF) {
		/* Expand our buffer if the line doesn't fit */
		if (len > linesz)
			buf = realloc(buf, linesz = (len + 4));

		rlen = fread(buf, 1, len, fp);
		buf[rlen - 1] = '\0';
		commentclr(buf);

		if (strlen(buf) == 0)
			continue;
		tmp = realloc(tmp, strlen(buf) + 1);
		res = sscanf(buf, " %d \"%[^\"]\" ", &index, tmp);

		if (res == 0)
			continue;

		if (index >= msgs)
			ma = realloc(ma, sizeof(char *) * (index + 1));

		if (res == 2) {
			ma[index] = malloc(strlen(tmp) + 1);
			strncpy(ma[index], tmp, strlen(tmp) + 1);
		}
		else {
			ma[index] = malloc(1);
			*ma[index] = '\0';
		}

		for (i = prev_index + 1; i < index; i++)
			ma[i] = NULL;

		msgs++;
		prev_index = index;
		if (index > high_index);
			high_index = index;
	}

	fclose(fp);
	if (msgs == 0 && ma != NULL)
		free(ma);
	if (tmp != NULL)
		free(tmp);

	*msg = ma;
	return (high_index);
}

/*
 * Free the message array returned from parse_msgfile
 */
void
parse_freemsg(char *msg[], int size)
{
	int i;

	for (i = 0; i < size; i++)
		if (msg[i] != NULL);
			free(msg[i]);
	free(msg);
}

/*
 * GUI callback for message translation.
 * If a message file was configured pass those messages to pam
 * for display.
 * This is suitable for graphical login managers such as GDM.
 */
BioAPI_RETURN
bioapi_guicallback(void *chptr, BioAPI_GUI_STATE guistate,
    BioAPI_GUI_RESPONSE resp, BioAPI_GUI_MESSAGE msg,
    BioAPI_GUI_PROGRESS progress, BioAPI_GUI_BITMAP_PTR Buffer)
{
	struct callbackh *ch = chptr;

	if (ch->prevmsg != msg && msg < ch->maxmsg) {
		ch->prevmsg = msg;
		if (ch->msgs[msg] != NULL) {
			pam_info(ch->pamh, "");
			pam_info(ch->pamh, "%s", ch->msgs[msg]);
		}
	}
	return (BioAPI_OK);
}

/*
 * authentication management
 */
PAM_EXTERN int
pam_sm_authenticate(pam_handle_t *pamh, int flags __unused,
    int argc, const char *argv[])
{
	const char *user, *bsp_id, *dbid, *conf, *msgfile;
	int error, pam_retval = PAM_AUTH_ERR;
	BioAPI_HANDLE *handle;
	struct birdb_rec keyrec, **recs;
	struct birdb_mod *bm;
	struct callbackh ch;
	uid_t euid;
	birdb *bdb;
	void *bmh;
	int opt;

	if (argc < 1) {
		PAM_VERBOSE_ERROR("BiAPI BSP UUID missing");
		return (PAM_SERVICE_ERR);
	}
	else if (argc < 2) {
		PAM_VERBOSE_ERROR("Missing backend module");
		return (PAM_SERVICE_ERR);
	}

	bsp_id = argv[0];
	PAM_LOG("Got BioAPI BSP UUID: %s", bsp_id);
	dbid = argv[1];
	PAM_LOG("Got BIRDB backend module: %s", dbid);

	conf = DEFCONFPATH; 
	msgfile = NULL;

	optind = 2;
	while ((opt = getopt(argc, (char **)argv, "m:f:")) != -1) {
		switch (opt) {
		case 'm':
			msgfile = argv[optind - 1];
			PAM_LOG("Got message callback: %s", msgfile);
			break;
		case 'f':
			conf = argv[optind - 1];
			PAM_LOG("Got birdb configuration file: %s", conf);
			break;
		}
	}

	euid = geteuid();
	if (euid != 0)
		return (PAM_PERM_DENIED);

	if (openpam_get_option(pamh, PAM_OPT_AUTH_AS_SELF)) {
		user = getlogin();
	}
	else {
		error = pam_get_user(pamh, &user, NULL);	
		if (error != PAM_SUCCESS)
			return (error);
	}
	PAM_LOG("Got user: %s", user);

	setuid(euid);
	pam_info(pamh, "Initiating biometric authentication..."); 

	error = bioapi_init();
	if (error)
		goto out;

	if (msgfile != NULL) {
		ch.pamh = pamh;
		ch.maxmsg = parse_msgfile(msgfile, &ch.msgs);
		ch.prevmsg = -1;
	}
	else {
		ch.maxmsg = 0;
	}

	bdb = birdb_init();
	if (bdb == NULL) {
		PAM_VERBOSE_ERROR("Failed to initate the BIR DB"); 
		goto out;
	}

	error = birdb_cfgparse(bdb, conf);
	if (error < 0) {
		PAM_VERBOSE_ERROR("Failed to parse %s", conf);
		goto out;
	}

	bm = birdb_findmod(bdb, dbid);
	if (bm == NULL) {
		PAM_VERBOSE_ERROR("Unable to find backend ``%s''", dbid);
		goto error1;
	}

	bmh = birdb_backend_open(bm, bsp_id, bm->bm_argc, bm->bm_argv);
	if (bmh == NULL) {
		PAM_VERBOSE_ERROR("Failed to initiate backend ``%s''", dbid);
		goto error1;
	}

	keyrec.br_key = (char *)user;
	recs = birdb_backend_get(bm, bmh, &keyrec);
	if (recs != NULL) {
		handle = bioapi_attach_bsp(bsp_id);
		if (handle == NULL) {
			PAM_VERBOSE_ERROR("Failed to attach the selected BSP");
			goto error2;
		}
		if (ch.maxmsg > 0)
			BioAPI_SetGUICallbacks(*handle, NULL, NULL,
			    bioapi_guicallback, &ch);
		error = bioapi_verify_many(handle, recs);
		PAM_LOG("Got biometric authentication");
		if (error >= 0)
			pam_retval = PAM_SUCCESS;
		else
			PAM_VERBOSE_ERROR("Biometric authentication refused");
		bioapi_detach_bsp(handle, bsp_id);
	}
	else {
		PAM_VERBOSE_ERROR("No biometric data avaiable");
	}

error2:
	birdb_backend_freegetres(bm, bmh, recs);
	birdb_backend_close(bm, bmh);
error1:
	birdb_close(bdb);
	bioapi_destroy();
out:
	if (msgfile != NULL)
		parse_freemsg(ch.msgs, ch.maxmsg);
	return (pam_retval);
}

PAM_EXTERN int
pam_sm_setcred(pam_handle_t *pamh __unused, int flags __unused,
    int argc __unused, const char *argv[] __unused)
{

    return (PAM_SUCCESS);
}

/*
 * "Change" biometric record
 * Instead of changing we just enroll another record, depending
 * on the backend this will create a new record. But backends
 * only supporing one record per user will treat this as a change.
 */
PAM_EXTERN int
pam_sm_chauthtok(pam_handle_t *pamh, int flags,
    int argc , const char *argv[])
{
	const char *user, *bsp_id, *dbid, *conf, *msgfile;
	int error, pam_retval = PAM_AUTH_ERR;
	BioAPI_HANDLE *handle;
	struct birdb_rec keyrec, *rec, **recs;
	struct birdb_mod *bm;
	struct callbackh ch;
	uid_t euid, ruid;
	birdb *bdb;
	void *bmh;
	int opt;

	if (argc < 1) {
		PAM_VERBOSE_ERROR("BiAPI BSP UUID missing");
		return (PAM_SERVICE_ERR);
	}
	else if (argc < 2) {
		PAM_VERBOSE_ERROR("Missing backend module");
		return (PAM_SERVICE_ERR);
	}

	bsp_id = argv[0];
	PAM_LOG("Got BioAPI BSP UUID: %s", bsp_id);
	dbid = argv[1];
	PAM_LOG("Got BIRDB backend module: %s", dbid);

	conf = DEFCONFPATH; 
	msgfile = NULL;

	optind = 2;
	while ((opt = getopt(argc, (char **)argv, "m:f:")) != -1) {
		switch (opt) {
		case 'm':
			msgfile = argv[optind - 1];
			PAM_LOG("Got message callback: %s", msgfile);
			break;
		case 'f':
			conf = argv[optind - 1];
			PAM_LOG("Got birdb configuration file: %s", conf);
			break;
		}
	}

	euid = geteuid();
	ruid = getuid();

	if (euid != 0)
		return (PAM_PERM_DENIED);

	if (openpam_get_option(pamh, PAM_OPT_AUTH_AS_SELF)) {
		user = getlogin();
	}
	else {
		error = pam_get_user(pamh, &user, NULL);	
		if (error != PAM_SUCCESS)
			return (error);
	}
	PAM_LOG("Got user: %s", user);

	setuid(euid);

	error = bioapi_init();
	if (error)
		goto out;

	if (msgfile != NULL) {
		ch.pamh = pamh;
		ch.maxmsg = parse_msgfile(msgfile, &ch.msgs);
		ch.prevmsg = -1;
	}
	else {
		ch.maxmsg = 0;
	}

	bdb = birdb_init();
	if (bdb == NULL) {
		PAM_VERBOSE_ERROR("Failed to initate the BIR DB"); 
		goto out;
	}

	error = birdb_cfgparse(bdb, conf);
	if (error < 0) {
		PAM_VERBOSE_ERROR("Failed to parse %s", conf);
		goto out;
	}

	bm = birdb_findmod(bdb, dbid);
	if (bm == NULL) {
		PAM_VERBOSE_ERROR("Unable to find backend ``%s''", dbid);
		goto error1;
	}

	bmh = birdb_backend_open(bm, bsp_id, bm->bm_argc, bm->bm_argv);
	if (bmh == NULL) {
		PAM_VERBOSE_ERROR("Failed to initiate backend ``%s''", dbid);
		goto error1;
	}


	if (flags & PAM_PRELIM_CHECK) {
		if (ruid == 0) {
			pam_retval = pam_set_data(pamh, "authtoken", NULL, NULL);
		}
		else {
			keyrec.br_key = (char *)user;
			recs = birdb_backend_get(bm, bmh, &keyrec);
			pam_set_data(pamh, "authtoken", recs, NULL);
			if (recs != NULL)
				pam_retval = PAM_SUCCESS; 
			else {
				PAM_VERBOSE_ERROR(
				    "Failed to obtain biometric data");
				pam_retval = PAM_ABORT;
			}
		}
	}
	else if (flags & PAM_UPDATE_AUTHTOK) {
		error = pam_get_data(pamh, "authtoken", (void **)&recs);
		if (error != PAM_SUCCESS)
			goto error2;

		handle = bioapi_attach_bsp(bsp_id);
		if (handle == NULL) {
			PAM_VERBOSE_ERROR("Failed to attach the selected BSP");
			goto error2;
		}
		if (ch.maxmsg > 0)
			BioAPI_SetGUICallbacks(*handle, NULL, NULL,
			    bioapi_guicallback, &ch);

		if (recs != NULL) {
			error = bioapi_verify_many(handle, recs);
			birdb_backend_freegetres(bm, bmh, recs);
		
			if (error < 0) {
				PAM_VERBOSE_ERROR(
				    "Biometric authentication refused");
				bioapi_detach_bsp(handle, bsp_id);
				if (msgfile != NULL)
					parse_freemsg(ch.msgs, ch.maxmsg);
				goto error2;	
			}
		}

		rec = bioapi_enroll(handle, bm, bmh, user);
		if (rec != NULL) {
			pam_info(pamh, "Please verify record..."); 

            error = bioapi_verify(handle, rec);
            if (error == 0) {
				pam_info(pamh, "Record created successfully\n");
				pam_retval = PAM_SUCCESS;
			}
            else {
                birdb_backend_del(bm, bmh, rec);
                PAM_VERBOSE_ERROR("Records do not match\n");
            }
        }
        else
            PAM_VERBOSE_ERROR("Failed to create BIR record\n");

        birdb_freerec(rec);
		bioapi_detach_bsp(handle, bsp_id);
		if (msgfile != NULL)
			parse_freemsg(ch.msgs, ch.maxmsg);
	}
	else {
		pam_retval = PAM_ABORT;
	}

error2:
	birdb_backend_close(bm, bmh);
error1:
	birdb_close(bdb);
	bioapi_destroy();
out:
	return (pam_retval);
}

PAM_EXTERN int
pam_sm_acct_mgmt(pam_handle_t *pamh, int flags __unused,
    int argc __unused, const char *argv[] __unused)
{

	return (PAM_SUCCESS);
}

PAM_MODULE_ENTRY("pam_bsdbioapi");

==> pam_bsdbioapi-1.5.1.tar.gz: pam_bsdbioapi-1.5.1/src/pam_bsdbioapi/pam_bsdbioapi.c <==
/*-
 * Copyright (c) 2006 Fredrik Lindberg. <fli at shapeshifter dot se>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: pam_bsdbioapi.c 62 2008-03-15 11:23:39Z fli $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/queue.h>
#include <login_cap.h>
#include <pwd.h>

#include <bioapi.h>
#include <birdb.h>
#include <int_bioapi.h>

#define PAM_SM_AUTH
#define PAM_SM_PASSWORD

#include <security/pam_appl.h>
#include <security/pam_modules.h>
#include <security/pam_mod_misc.h>

#define DEFCONFPATH "/usr/local/etc/birdb.conf"

/*
 * Callback state, passed to bioapi_guicallback
 */
struct callbackh {
	pam_handle_t *pamh;
	char **msgs;
	int maxmsg;
	int prevmsg;
};

/*
 * Return position and length of next logical line (separated by \n)
 */
static int
nextline(FILE *fp, size_t *len)
{
	int ch;
	long tmp, newlen = 0;

	/* Save our file position */
	tmp = ftell(fp);

	/* Seek until we hit \n (end of line) or EOF */
	do {
#if __POSIX_VISIBLE >= 199506
		ch = getc_unlocked(fp);
#else
		ch = getc(fp);
#endif
		newlen++;
	} while ((ch != EOF) && ((char)ch != '\n'));

	/* Restore position to previous value */
	fseek(fp, tmp, SEEK_SET);

	*len = newlen;
	return (ch);
}

/*
 * Truncate string at the occurance of a '#'-character
 */
static void
commentclr(char *str)
{
	char *p = str;

	do {
		if (*p == '#') { *p = 0; }
	} while (*p++ != 0);
}

/*
 * Parse callback message files
 * Callers must cope with that the returned array may have holes in
 * it, the holes does always contain a NULL pointer.
 */
static int
parse_msgfile(const char *path, char **msg[])
{
	size_t len, linesz = 80, rlen, res;
	char *buf, *tmp = NULL, **ma = NULL;
	int msgs = 0, index, prev_index, high_index, i;
	FILE *fp;

	fp = fopen(path, "r");
	if (fp == NULL)
		return (-1);

	buf = malloc(linesz);
	if (buf == NULL)
		return (-1);

	prev_index = high_index = 0;
	ma = malloc(sizeof(char *));
	while (nextline(fp, &len) != EOF) {
		/* Expand our buffer if the line doesn't fit */
		if (len > linesz)
			buf = realloc(buf, linesz = (len + 4));

		rlen = fread(buf, 1, len, fp);
		buf[rlen - 1] = '\0';
		commentclr(buf);

		if (strlen(buf) == 0)
			continue;
		tmp = realloc(tmp, strlen(buf) + 1);
		res = sscanf(buf, " %d \"%[^\"]\" ", &index, tmp);

		if (res == 0)
			continue;

		if (index >= msgs)
			ma = realloc(ma, sizeof(char *) * (index + 1));

		if (res == 2) {
			ma[index] = malloc(strlen(tmp) + 1);
			strncpy(ma[index], tmp, strlen(tmp) + 1);
		}
		else {
			ma[index] = malloc(1);
			*ma[index] = '\0';
		}

		for (i = prev_index + 1; i < index; i++)
			ma[i] = NULL;

		msgs++;
		prev_index = index;
		if (index > high_index);
			high_index = index;
	}

	fclose(fp);
	if (msgs == 0 && ma != NULL)
		free(ma);
	if (tmp != NULL)
		free(tmp);

	*msg = ma;
	return (high_index);
}

/*
 * Free the message array returned from parse_msgfile
 */
void
parse_freemsg(char *msg[], int size)
{
	int i;

	for (i = 0; i < size; i++)
		if (msg[i] != NULL);
			free(msg[i]);
	free(msg);
}

/*
 * GUI callback for message translation.
 * If a message file was configured pass those messages to pam
 * for display.
 * This is suitable for graphical login managers such as GDM.
 */
BioAPI_RETURN
bioapi_guicallback(void *chptr, BioAPI_GUI_STATE guistate,
    BioAPI_GUI_RESPONSE resp, BioAPI_GUI_MESSAGE msg,
    BioAPI_GUI_PROGRESS progress, BioAPI_GUI_BITMAP_PTR Buffer)
{
	struct callbackh *ch = chptr;

	if (ch->prevmsg != msg && msg >= 0 && msg < ch->maxmsg) {
		ch->prevmsg = msg;
		if (ch->msgs[msg] != NULL) {
			pam_info(ch->pamh, "");
			pam_info(ch->pamh, "%s", ch->msgs[msg]);
		}
	}
	return (BioAPI_OK);
}

/*
 * authentication management
 */
PAM_EXTERN int
pam_sm_authenticate(pam_handle_t *pamh, int flags __unused,
    int argc, const char *argv[])
{
	const char *user, *bsp_id, *dbid, *conf, *msgfile;
	int error, pam_retval = PAM_AUTH_ERR, skip_unenrolled;
	BioAPI_HANDLE *handle;
	struct birdb_rec keyrec, **recs;
	struct birdb_mod *bm;
	struct callbackh ch;
	uid_t euid;
	birdb *bdb;
	void *bmh;
	int opt;

	if (argc < 1) {
		PAM_VERBOSE_ERROR("BiAPI BSP UUID missing");
		return (PAM_SERVICE_ERR);
	}
	else if (argc < 2) {
		PAM_VERBOSE_ERROR("Missing backend module");
		return (PAM_SERVICE_ERR);
	}

	bsp_id = argv[0];
	PAM_LOG("Got BioAPI BSP UUID: %s", bsp_id);
	dbid = argv[1];
	PAM_LOG("Got BIRDB backend module: %s", dbid);

	conf = DEFCONFPATH;
	msgfile = NULL;
	skip_unenrolled = 0;

	optind = 2;
	while ((opt = getopt(argc, (char **)argv, "m:f:s")) != -1) {
		switch (opt) {
		case 'm':
			msgfile = argv[optind - 1];
			PAM_LOG("Got message callback: %s", msgfile);
			break;
		case 'f':
			conf = argv[optind - 1];
			PAM_LOG("Got birdb configuration file: %s", conf);
			break;
		case 's':
			skip_unenrolled = 1;
			break;
		}
	}

	euid = geteuid();
	if (euid != 0)
		return (PAM_PERM_DENIED);

	if (openpam_get_option(pamh, PAM_OPT_AUTH_AS_SELF)) {
		user = getlogin();
	}
	else {
		error = pam_get_user(pamh, &user, NULL);	
		if (error != PAM_SUCCESS)
			return (error);
	}
	PAM_LOG("Got user: %s", user);

	setuid(euid);

	error = bioapi_init();
	if (error)
		goto out;

	if (msgfile != NULL) {
		ch.pamh = pamh;
		ch.maxmsg = parse_msgfile(msgfile, &ch.msgs);
		ch.prevmsg = -1;
	}
	else {
		ch.maxmsg = 0;
	}

	bdb = birdb_init();
	if (bdb == NULL) {
		PAM_VERBOSE_ERROR("Failed to initate the BIR DB");
		goto out;
	}

	error = birdb_cfgparse(bdb, conf);
	if (error < 0) {
		PAM_VERBOSE_ERROR("Failed to parse %s", conf);
		goto out;
	}

	bm = birdb_findmod(bdb, dbid);
	if (bm == NULL) {
		PAM_VERBOSE_ERROR("Unable to find backend ``%s''", dbid);
		goto error1;
	}

	bmh = birdb_backend_open(bm, bsp_id, bm->bm_argc, bm->bm_argv);
	if (bmh == NULL) {
		PAM_VERBOSE_ERROR("Failed to initiate backend ``%s''", dbid);
		goto error1;
	}

	keyrec.br_key = (char *)user;
	recs = birdb_backend_get(bm, bmh, &keyrec);
	if (recs != NULL && (!skip_unenrolled || recs[0] != NULL)) {
		pam_info(pamh, "Initiating biometric authentication...");
		handle = bioapi_attach_bsp(bsp_id);
		if (handle == NULL) {
			PAM_VERBOSE_ERROR("Failed to attach the selected BSP");
			goto error2;
		}
		if (ch.maxmsg > 0)
			BioAPI_SetGUICallbacks(*handle, NULL, NULL,
			    (BioAPI_GUI_STATE_CALLBACK)bioapi_guicallback, &ch);
		error = bioapi_verify_many(handle, recs);
		PAM_LOG("Got biometric authentication");
		if (error >= 0)
			pam_retval = PAM_SUCCESS;
		else
			PAM_VERBOSE_ERROR("Biometric authentication refused");
		bioapi_detach_bsp(handle, bsp_id);
	}
	else {
		PAM_VERBOSE_ERROR("No biometric data avaiable");
	}

error2:
	birdb_backend_freegetres(bm, bmh, recs);
	birdb_backend_close(bm, bmh);
error1:
	birdb_close(bdb);
	bioapi_destroy();
out:
	if (msgfile != NULL)
		parse_freemsg(ch.msgs, ch.maxmsg);
	return (pam_retval);
}

PAM_EXTERN int
pam_sm_setcred(pam_handle_t *pamh __unused, int flags __unused,
    int argc __unused, const char *argv[] __unused)
{

    return (PAM_SUCCESS);
}

/*
 * "Change" biometric record
 * Instead of changing we just enroll another record, depending
 * on the backend this will create a new record. But backends
 * only supporing one record per user will treat this as a change.
 */
PAM_EXTERN int
pam_sm_chauthtok(pam_handle_t *pamh, int flags,
    int argc , const char *argv[])
{
	const char *user, *bsp_id, *dbid, *conf, *msgfile;
	int error, pam_retval = PAM_AUTH_ERR;
	BioAPI_HANDLE *handle;
	struct birdb_rec keyrec, *rec, **recs;
	struct birdb_mod *bm;
	struct callbackh ch;
	uid_t euid, ruid;
	birdb *bdb;
	void *bmh;
	int opt;

	if (argc < 1) {
		PAM_VERBOSE_ERROR("BiAPI BSP UUID missing");
		return (PAM_SERVICE_ERR);
	}
	else if (argc < 2) {
		PAM_VERBOSE_ERROR("Missing backend module");
		return (PAM_SERVICE_ERR);
	}

	bsp_id = argv[0];
	PAM_LOG("Got BioAPI BSP UUID: %s", bsp_id);
	dbid = argv[1];
	PAM_LOG("Got BIRDB backend module: %s", dbid);

	conf = DEFCONFPATH;
	msgfile = NULL;

	optind = 2;
	while ((opt = getopt(argc, (char **)argv, "m:f:")) != -1) {
		switch (opt) {
		case 'm':
			msgfile = argv[optind - 1];
			PAM_LOG("Got message callback: %s", msgfile);
			break;
		case 'f':
			conf = argv[optind - 1];
			PAM_LOG("Got birdb configuration file: %s", conf);
			break;
		}
	}

	euid = geteuid();
	ruid = getuid();

	if (euid != 0)
		return (PAM_PERM_DENIED);

	if (openpam_get_option(pamh, PAM_OPT_AUTH_AS_SELF)) {
		user = getlogin();
	}
	else {
		error = pam_get_user(pamh, &user, NULL);	
		if (error != PAM_SUCCESS)
			return (error);
	}
	PAM_LOG("Got user: %s", user);

	setuid(euid);

	error = bioapi_init();
	if (error)
		goto out;

	if (msgfile != NULL) {
		ch.pamh = pamh;
		ch.maxmsg = parse_msgfile(msgfile, &ch.msgs);
		ch.prevmsg = -1;
	}
	else {
		ch.maxmsg = 0;
	}

	bdb = birdb_init();
	if (bdb == NULL) {
		PAM_VERBOSE_ERROR("Failed to initate the BIR DB");
		goto out;
	}

	error = birdb_cfgparse(bdb, conf);
	if (error < 0) {
		PAM_VERBOSE_ERROR("Failed to parse %s", conf);
		goto out;
	}

	bm = birdb_findmod(bdb, dbid);
	if (bm == NULL) {
		PAM_VERBOSE_ERROR("Unable to find backend ``%s''", dbid);
		goto error1;
	}

	bmh = birdb_backend_open(bm, bsp_id, bm->bm_argc, bm->bm_argv);
	if (bmh == NULL) {
		PAM_VERBOSE_ERROR("Failed to initiate backend ``%s''", dbid);
		goto error1;
	}


	if (flags & PAM_PRELIM_CHECK) {
		if (ruid == 0) {
			pam_retval = pam_set_data(pamh, "authtoken", NULL, NULL);
		}
		else {
			keyrec.br_key = (char *)user;
			recs = birdb_backend_get(bm, bmh, &keyrec);
			pam_set_data(pamh, "authtoken", recs, NULL);
			if (recs != NULL)
				pam_retval = PAM_SUCCESS;
			else {
				PAM_VERBOSE_ERROR(
				    "Failed to obtain biometric data");
				pam_retval = PAM_ABORT;
			}
		}
	}
	else if (flags & PAM_UPDATE_AUTHTOK) {
		error = pam_get_data(pamh, "authtoken", (void **)&recs);
		if (error != PAM_SUCCESS)
			goto error2;

		handle = bioapi_attach_bsp(bsp_id);
		if (handle == NULL) {
			PAM_VERBOSE_ERROR("Failed to attach the selected BSP");
			goto error2;
		}
		if (ch.maxmsg > 0)
			BioAPI_SetGUICallbacks(*handle, NULL, NULL,
			    (BioAPI_GUI_STATE_CALLBACK)bioapi_guicallback, &ch);

		if (recs != NULL) {
			error = bioapi_verify_many(handle, recs);
			birdb_backend_freegetres(bm, bmh, recs);
		
			if (error < 0) {
				PAM_VERBOSE_ERROR(
				    "Biometric authentication refused");
				bioapi_detach_bsp(handle, bsp_id);
				if (msgfile != NULL)
					parse_freemsg(ch.msgs, ch.maxmsg);
				goto error2;	
			}
		}

		rec = bioapi_enroll(handle, bm, bmh, user);
		if (rec != NULL) {
			pam_info(pamh, "Please verify record...");

            error = bioapi_verify(handle, rec);
            if (error == 0) {
				pam_info(pamh, "Record created successfully\n");
				pam_retval = PAM_SUCCESS;
			}
            else {
                birdb_backend_del(bm, bmh, rec);
                PAM_VERBOSE_ERROR("Records do not match\n");
            }
        }
        else
            PAM_VERBOSE_ERROR("Failed to create BIR record\n");

        birdb_freerec(rec);
		bioapi_detach_bsp(handle, bsp_id);
		if (msgfile != NULL)
			parse_freemsg(ch.msgs, ch.maxmsg);
	}
	else {
		pam_retval = PAM_ABORT;
	}

error2:
	birdb_backend_close(bm, bmh);
error1:
	birdb_close(bdb);
	bioapi_destroy();
out:
	return (pam_retval);
}

PAM_EXTERN int
pam_sm_acct_mgmt(pam_handle_t *pamh, int flags __unused,
    int argc __unused, const char *argv[] __unused)
{

	return (PAM_SUCCESS);
}

PAM_MODULE_ENTRY("pam_bsdbioapi");
